
#define FROM_ASM
#include "rtos.h"


current_time = TIMER_REGISTER
the_block_size_is = BLOCK_SIZE
the_num_blocks_is = NUM_BLOCKS

# defined here
.global reset
.global call_function
.global exit
.global _exit
.global memmove
.global memcpy
.global print
.global current_time
.global high_water
.global notify_task_name

# defined elsewhere
.global main

	.section	.kernelstuff.t,"ax",@progbits
.org 0x0
reset:
        # SYSTEM mode: interrupts off, breaks on
        msrset  r3, MSR_BE
        bri     kernel_entry
.org 0x8
exception:
        brai    exception       # crash
.org 0x10
interrupt:
        brai    interrupt       # crash
.org 0x18
break:
        brai    break           # crash
.org 0x20
hw_exception:
        brai    hw_exception    # crash

.org 0x50
kernel_entry:
        # Kernel entry point, set up globals
        addik   r1, r0, SYSTEM_STACK

        # Initialise C runtime
        addi	r6, r0, __bss_start             /* clear BSS */
        addi	r7, r0, __bss_end		
        rsub	r18, r6, r7		
        blei	r18, .Lendbss
.Lloopbss:	
        swi	r0, r6, 0
        addi	r6, r6, 4
        rsub	r18, r6, r7
        bgti	r18, .Lloopbss
.Lendbss:

        addi    r5, r0, 0
        addi    r6, r0, 0

        brlid	r15, main
        addi    r7, r0, 0

_exit:
exit:   
        # The following are needed to make our simulator happy; they
        # pretend to be the RTOS kernel
        swi     r0, r0, current_time

        # SYSCALL mode: disable interrupts & breaks
        msrclr  r3, MSR_IE | MSR_BE

0:      .long   0x80000001
        bri     0b          # exit loop
       
call_function:
        # r5: address of target function
        # r6: block info word
        # r7-r10: parameters of target function

        # register saving on stack
        addik   r1, r1, -16
        swi     r15, r1, 0
        swi     r19, r1, 4
        lwi     r3, r0, high_water
        swi     r3, r1, 8

        # stack is moved to match real kernel
        ori     r19, r1, 0
        andni   r1, r1, BLOCK_SIZE - 1  # round r1 down

        # notify simulator
        addik   r3, r0, TRC_CAPT_CALL_FUNCTION
        put     r5, rfsl0   
        put     r1, rfsl1   
        cput    r3, rfsl4 

        swi     r0, r0, high_water

        # Parameter registers moved into place
        or      r4, r7, r0
        or      r6, r8, r0
        or      r7, r9, r0
        or      r8, r10, r0

        # Call function
        brald   r15, r5
        or      r5, r4, r0

return_point:

        addik   r6, r0, TRC_CAPT_RETURN_FUNCTION
        lwi     r5, r0, high_water
        put     r5, rfsl0   
        put     r1, rfsl1   
        cput    r6, rfsl4 

        ori     r1, r19, 0
        lwi     r15, r1, 0
        lwi     r19, r1, 4
        lwi     r5, r1, 8

        swi     r5, r0, high_water

        rtsd    r15, 8
        addik   r1, r1, 16
        

memcpy:
memmove: # r5 is dest
        # r6 is source
        # r7 is size

        # DMA(r6 to r5)
        put     r6, rfsl0        # src
        put     r5, rfsl1        # dest
        cput    r7, rfsl0        # size

        rtsd    r15, 8
        or      r3, r5, r0


print:  # r5 is address of 0-terminated string
        rtsd    r15, 8
        cput    r5, rfsl2

notify_task_name:
        # r5 = task name
        # r6 = function ptr
        put     r6, rfsl0       # function
        rtsd    r15, 8
        cput    r5, rfsl5       # task name

.data
high_water:
        .long   0

