/* RTOS low-level code */
/* Cache variant */

#define FROM_ASM
#include "rtos.h"


rmsr_offset = 0x00
r1_offset = 0x04
r2_offset = 0x08
r3_offset = 0x0c
r4_offset = 0x10
r5_offset = 0x14
r6_offset = 0x18
r7_offset = 0x1c
r8_offset = 0x20
r9_offset = 0x24
r10_offset = 0x28
r11_offset = 0x2c
r12_offset = 0x30
r13_offset = 0x34
r14_offset = 0x38
r15_offset = 0x3c
r15_b_offset = 0x40
r18_offset = 0x44
r19_offset = 0x48
r1_b_offset = 0x4c
switch_area_size = r1_b_offset + 4
current_time = TIMER_REGISTER


# defined here
.global reset
.global call_function
.global exit
.global _exit
.global memmove
.global memcpy
.global _open
.global print
.global notify_task_name
.global current_time
.global kernel_idle

# defined elsewhere
.global scheduler
.global main

#define MSR_CACHES 0xa0

	.section	.kernelstuff.t,"ax",@progbits
.org 0x0
reset:
        # SYSTEM mode: interrupts off, breaks on, activate caches
        msrset  r3, MSR_BE | MSR_CACHES
        bri     kernel_entry
.org 0x8
exception:
        brai    exception       # crash
.org 0x10
interrupt:
        brai    interrupt_handler
.org 0x18
break:
        brai    break           # crash
.org 0x20
hw_exception:
        brai    hw_exception    # crash

.org 0x50
.align 2
kernel_entry:
        # Kernel entry point, set up globals
        addik   r1, r0, SYSTEM_STACK
        addik   r17, r0, 0xdeadbeef /* not used in cache version */

        # Initialise C runtime
        addi	r6, r0, __bss_start             /* clear BSS */
        addi	r7, r0, __bss_end		
        rsub	r18, r6, r7		
        blei	r18, .Lendbss
.Lloopbss:	
        swi	r0, r6, 0
        addi	r6, r6, 4
        rsub	r18, r6, r7
        bgti	r18, .Lloopbss
.Lendbss:

        addi    r5, r0, 0
        addi    r6, r0, 0

        brlid	r15, main
        addi    r7, r0, 0

_exit:
exit:   # SYSTEM mode: disable interrupts
        msrclr  r3, MSR_IE 
        addik   r3, r0, exit_called
        cput    r3, rfsl2
0:      .long   0x80000001
        bri     0b          # exit loop
       
task_func_is_zero:
        addik   r5, r0, task_func_is_zero_called
        cput    r5, rfsl2
        bri     exit
    
interrupt_handler:
        # Save registers not saved by call_function
        addik   r1, r1, -switch_area_size
        swi     r3, r1, r3_offset
        swi     r4, r1, r4_offset
        swi     r5, r1, r5_offset
        swi     r6, r1, r6_offset
        swi     r7, r1, r7_offset
        swi     r8, r1, r8_offset
        swi     r9, r1, r9_offset
        swi     r10,r1, r10_offset
        swi     r11,r1, r11_offset
        swi     r12,r1, r12_offset
        swi     r14,r1, r14_offset
        swi     r15,r1, r15_b_offset
        swi     r18,r1, r18_offset
        swi     r19,r1, r19_offset

        # Where shall we call next?
        brlid   r15, scheduler
        nop

        or      r19, r3, r0     # TCB pointer copied to non-volatile
        beqid   r3, do_not_task_switch
        or      r5, r0, r0
        lwi     r5, r3, 0       # get task function

        # Notify: a new task will begin
        beqid   r5, task_func_is_zero
        addik   r4, r0, TRC_TASK_START
        put     r5, rfsl0
        cput    r4, rfsl4 

        # USER mode: enable interrupts, breaks
        msrset  r3, MSR_IE | MSR_BE

        # Call the task
        brald   r15, r5
        nop

        # SYSTEM mode: disable interrupts
        msrclr  r5, MSR_IE 

        # Task exit
        swi     r0, r19, 4      # clear running flag
        lwi     r5, r19, 0      # get task function

        # Preempt immediately on return from ISR in case we have 
        # another task waiting to run
        swi     r0, r0, current_time

        # Notify: task finished
        addik   r4, r0, TRC_TASK_END
        put     r5, rfsl0
        cput    r4, rfsl4 

do_not_task_switch:
        # Notify: returning from interrupt
        addik   r4, r0, TRC_IN_RTI
        put     r5, rfsl0
        cput    r4, rfsl4 

        lwi     r3, r1, r3_offset
        lwi     r4, r1, r4_offset
        lwi     r5, r1, r5_offset
        lwi     r6, r1, r6_offset
        lwi     r7, r1, r7_offset
        lwi     r8, r1, r8_offset
        lwi     r9, r1, r9_offset
        lwi     r10,r1, r10_offset
        lwi     r11,r1, r11_offset
        lwi     r12,r1, r12_offset
        lwi     r14,r1, r14_offset
        lwi     r15,r1, r15_b_offset
        lwi     r18,r1, r18_offset
        lwi     r19,r1, r19_offset

        # Return from interrupt
        rtid    r14, 0
        addi    r1, r1, switch_area_size

call_function:
        # r5: address of target function
        # r6: block info word
        # r7-r10: parameters of target function
        # r15: return PC - 8
        or      r4, r7, r0
        or      r6, r8, r0
        or      r7, r9, r0
        or      r8, r10, r0

        # Call function (chain)
        brad    r5
        or      r5, r4, r0

memcpy:
memmove: # r5 is dest
        # r6 is source
        # r7 is size

        # SYSCALL mode: disable interrupts, breaks
        msrclr  r3, MSR_IE | MSR_BE

        # DMA(r6 to r5)
        put     r6, rfsl0        # src
        put     r5, rfsl1        # dest
        cput    r7, rfsl0        # size

        or      r3, r5, r0
        rtsd    r15, 8
        # USER mode: enable interrupts, breaks
        msrset  r5, MSR_IE | MSR_BE

print:  # r5 is address of 0-terminated string
        rtsd    r15, 8
        cput    r5, rfsl2

notify_task_name:
        # r5 = task name
        # r6 = function ptr
        put     r6, rfsl0       # function
        rtsd    r15, 8
        cput    r5, rfsl5       # task name

_open:  rtsd    r15, 8
        nop

kernel_idle:
        # Notify: idle task start
        addik   r5, r0, TRC_TASK_START
        put     r0, rfsl0
        cput    r5, rfsl4 

        # Clock start - preempt as soon as interrupts are enabled
        # The clock has to be enabled in SYSTEM mode for the sake of the 
        # simulator which assumes that stopped clock == SYSTEM.
        swi     r0, r0, current_time

        # USER mode: enable interrupts, breaks
        msrset  r5, MSR_IE | MSR_BE

        # Idle loop
0:      addi    r5, r0, idle_called
        cput    r5, rfsl2 # print
        cput    r5, rfsl3 # HALT
        bri     0b

.data
exit_called:
        .ascii  "exit called"
        .byte   10
        .byte   0
idle_called:
        .ascii  "idle"
        .byte   10
        .byte   0
task_func_is_zero_called:
        .ascii  "Task function pointer is zero!"
        .byte   10
        .byte   0


